''
'' MikeHawk, 2021 - Use at your own risk.
'' Search for a file in the working directory and all of its sub-directories.
'' v.1
''

'$INCLUDE: 'QB.BI'
DEFINT A-Z

DECLARE SUB searchFile (where AS STRING, query AS STRING)
DECLARE SUB getObjects (where AS STRING, query AS STRING)
DECLARE FUNCTION scopeDown% ()
DECLARE FUNCTION scopeUp$ ()
DECLARE FUNCTION scopeAppend% (folder AS STRING)
DECLARE FUNCTION INSTRREV% (source AS STRING, find AS STRING)

CONST moveROOT = 0
CONST moveDOWN = -1
CONST moveUP = 1

TYPE scopeInfo          ' OFS SZE   DESCRIPTION
	start AS INTEGER      ' 000 ..2   Offset of first folder in pathLst
	limit AS INTEGER      ' 002 ..2   Offset of last folder in pathLst
	entry AS INTEGER      ' 004 ..2   Offset of last visited folder
END TYPE                ' 6 BYTES - Scope descriptor

DIM SHARED pathLst AS STRING * 20000      ' Parent and current sub-folders
DIM SHARED pathOfs AS INTEGER             ' Writing offset for pathLst
DIM SHARED levlNfo(0 TO 63) AS scopeInfo  ' Root plus 63 levels deep
DIM SHARED levlNow AS INTEGER             ' How deep we are
DIM SHARED result AS STRING               ' Result string

DIM findMe AS STRING

findMe = COMMAND$
IF LEN(findMe) THEN
	searchFile "./", findMe
ELSE
	PRINT "USAGE: FindMe <filename>"
END IF

''
'' Get all objects in current directory
''
SUB getObjects (where AS STRING, query AS STRING)
	DIM DTA AS STRING * 44, MaskZ AS STRING
	DIM regs AS RegTypeX, objName AS STRING

	'' SETUP DTA SO WE DON'T DESTROY COMMAND$ ''
	regs.ax = &H1A00                  ' Set DTA function
	regs.dx = VARPTR(DTA)             ' DS:DX points to our DTA
	regs.ds = -1                      ' Use current value for DS
	CALL INTERRUPTX(&H21, regs, regs) ' Do the interrupt

	MaskZ = where + "*.*" + CHR$(0)   '  Mask (search all)
	regs.ax = &H4E00                  '  FindFirst
	regs.cx = 16                      '  Get all object types
	regs.dx = SADD(MaskZ)             '  DS:DX points to ASCIIZ file mask
	regs.ds = -1                      '  Use current DS

	'' PARSE ALL OBJECTS ''
	DO
		CALL INTERRUPTX(&H21, regs, regs)    ' Do the interrupt
		IF (regs.flags AND &H1) THEN EXIT DO ' No object left
		objName = UCASE$(MID$(DTA, 31, INSTR(31, DTA, CHR$(0)) - 31))
		' Folder, append to scope
		IF (ASC(MID$(DTA, &H15 + 1, 1)) AND &H10) THEN
			IF ((objName <> ".") AND (objName <> "..")) THEN
				IF scopeAppend%(objName) THEN PRINT "Buffer overflow!": END
			END IF
		' File, compare with query
		ELSE
			IF (objName = query) THEN result = result + CHR$(LEN(where)) + UCASE$(where)
		END IF
		' FindNext
		regs.ax = &H4F00
	LOOP
END SUB

''
'' Last occurence of a string
''
FUNCTION INSTRREV% (source AS STRING, find AS STRING)
	DIM ofs AS INTEGER
	DO
		INSTRREV% = ofs
		ofs = INSTR(ofs + 1, source, find)
	LOOP WHILE ofs
END FUNCTION

''
'' Append subfolder to scope. One byte is reserved to provide the length in
'' bytes of the folder name. This takes less memory than assuming that all
'' folders have 12-byte long names. This function returns -1 if the memory
'' buffer is saturated. Returns 0 if the sub-folder was successfully added.
''
FUNCTION scopeAppend% (folder AS STRING)
	DIM tmp AS STRING

	IF (levlNfo(levlNow).start = 0) THEN
		levlNfo(levlNow).start = pathOfs
		levlNfo(levlNow).entry = pathOfs
	END IF
	levlNfo(levlNow).limit = pathOfs

	tmp = LTRIM$(RTRIM$(UCASE$(folder)))
	tmp = CHR$(LEN(tmp)) + tmp
	IF ((pathOfs + LEN(tmp)) >= LEN(pathLst)) THEN
		scopeAppend% = -1
	ELSE
		MID$(pathLst, pathOfs, LEN(tmp)) = tmp
		pathOfs = pathOfs + LEN(tmp)
		scopeAppend% = 0
	END IF
END FUNCTION

''
'' Free current scope. Freeing means that we no longer need the sub-folder
'' list for this scope so we can rewrite it. We also clear the descriptor so
'' it can be re-used. This function returns -1 if root as been reached.
''
FUNCTION scopeDown%
	IF levlNfo(levlNow).start THEN pathOfs = levlNfo(levlNow).start
	levlNfo(levlNow).start = 0
	levlNfo(levlNow).limit = 0
	levlNfo(levlNow).entry = 0
	scopeDown% = (levlNow = 0)
END FUNCTION

''
'' scopeUp$ returns the next sub-folder in this scope we should be browsing.
'' If no folder is available, scopeUp$ will return a blank string and you
'' should scopeDown% instead.
''
FUNCTION scopeUp$
	DIM offset AS INTEGER, length AS INTEGER

	' Nothing left to see here
	IF (levlNfo(levlNow).entry > levlNfo(levlNow).limit) OR (levlNfo(levlNow).entry = 0) THEN
		scopeUp$ = ""
	' Get next folder in line to browse
	ELSE
		offset = levlNfo(levlNow).entry
		length = ASC(MID$(pathLst, offset, 1))
		offset = offset + length + 1
		scopeUp$ = MID$(pathLst, offset - length, length)
		levlNfo(levlNow).entry = offset
	END IF
END FUNCTION

''
'' <where> is the base directory (where we start looking for the file,) it is
'' ideally an absolute path and should be terminated with a forward slash.
'' <query> is the name of the file we're looking for.
''
SUB searchFile (where AS STRING, query AS STRING)
	DIM pathNxt AS STRING, pathAll AS STRING, move AS INTEGER
	DIM resOfs AS INTEGER, resLen AS INTEGER, resCnt AS INTEGER
	DIM tmr AS DOUBLE, spin AS INTEGER, maxScope AS INTEGER, maxBuffer AS INTEGER

	pathOfs = 1     ' Always set to 1 before starting
	move = moveROOT ' Pretend we're moving up (for folder list)
	levlNow = 0     ' We start at level 0
	tmr = TIMER     ' For benchmark

	pathAll = LTRIM$(RTRIM$(UCASE$(where)))
	IF (RIGHT$(pathAll, 1) <> "/") THEN pathAll = pathAll + "/"
	query = LTRIM$(RTRIM$(UCASE$(query)))

	DO
		IF (move <> moveDOWN) THEN
			CLS
			PRINT "Searching... "; MID$("-\|/", 1 + spin \ 4, 1)
			spin = ((spin + 1) AND &HF)

			' Append directories, search for file
			IF (move) THEN levlNow = levlNow + 1
			CALL getObjects(pathAll, query)
			IF (levlNow > maxScope) THEN maxScope = levlNow
			IF (pathOfs > maxBuffer) THEN maxBuffer = pathOfs
		END IF

		' Enter next sub-directory
		pathNxt = scopeUp$
		IF LEN(pathNxt) THEN
			pathAll = pathAll + pathNxt + "/"
			move = moveUP
		ELSE
			IF (scopeDown%) THEN EXIT DO
			levlNow = levlNow - 1
			pathAll = LEFT$(pathAll, INSTRREV%(LEFT$(pathAll, LEN(pathAll) - 1), "/"))
			move = moveDOWN
		END IF
	LOOP UNTIL LEN(INKEY$)

	PRINT
	IF LEN(result) THEN
		resOfs = 1
		DO
			resLen = ASC(MID$(result, resOfs, 1))
			PRINT CHR$(34); MID$(result, resOfs + 1, resLen); CHR$(34)
			resCnt = resCnt + 1
			resOfs = resOfs + 1 + resLen
		LOOP WHILE (resOfs < LEN(result))
	END IF
	PRINT LTRIM$(STR$(resCnt)); " file(s) with the name "; CHR$(34); query; CHR$(34); " found"
	PRINT
	PRINT "   Time elapsed:"; TIMER - tmr; "seconds"
	PRINT "Max buffer size:"; maxBuffer; "/20,000 bytes"
	PRINT "  Deepest scope:"; maxScope; "/63"
END SUB
